@@ -3,23 +3,28 @@
  *
  * Written by Mark Spencer <markster@digium.com>
  *
- * Copyright (C) 2001-2005, Digium
+ * Copyright (C) 2001-2005, Digium, Inc.
  * All Rights Reserved.
+ */
+
+/*
+ * See http://www.asterisk.org for more information about
+ * the Asterisk project. Please do not directly contact
+ * any of the maintainers of this project for assistance;
+ * the project provides a web site, mailing lists and IRC
+ * channels for your use.
  *
- * This program is free software; you can redistribute it and/or modify
- * it under the terms of the GNU General Public License as published by
- * the Free Software Foundation; either version 2 of the License, or
- * (at your option) any later version.
- * 
- * This program is distributed in the hope that it will be useful,
- * but WITHOUT ANY WARRANTY; without even the implied warranty of
- * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
- * GNU General Public License for more details.
- * 
- * You should have received a copy of the GNU General Public License
- * along with this program; if not, write to the Free Software
- * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA. 
+ * This program is free software, distributed under the terms of
+ * the GNU General Public License Version 2 as published by the
+ * Free Software Foundation. See the LICENSE file included with
+ * this program for more details.
  *
+ * In addition, when this program is distributed with Asterisk in
+ * any form that would qualify as a 'combined work' or as a
+ * 'derivative work' (but not mere aggregation), you can redistribute
+ * and/or modify the combination under the terms of the license
+ * provided with that copy of Asterisk, instead of the license
+ * terms granted here.
  */
 
 #include <unistd.h>
@@ -81,7 +86,7 @@ char *pri_switch2str(int sw)
 
 static void pri_default_timers(struct pri *pri, int switchtype)
 {
-	int defaulttimers[20][PRI_MAX_TIMERS] = PRI_TIMERS_ALL;
+	static const int defaulttimers[20][PRI_MAX_TIMERS] = PRI_TIMERS_ALL;
 	int x;
 
 	for (x = 0; x<PRI_MAX_TIMERS; x++) {
@@ -187,63 +192,76 @@ static int __pri_write(struct pri *pri, void *buf, int buflen)
 	return res;
 }
 
-static struct pri *__pri_new(int fd, int node, int switchtype, struct pri *master, pri_io_cb rd, pri_io_cb wr, void *userdata)
+/* Pass in the master for this function */
+void __pri_free_tei(struct pri * p)
+{
+	free (p);
+}
+
+struct pri *__pri_new_tei(int fd, int node, int switchtype, struct pri *master, pri_io_cb rd, pri_io_cb wr, void *userdata, int tei, int bri)
 {
 	struct pri *p;
-	p = malloc(sizeof(struct pri));
-	if (p) {
-		memset(p, 0, sizeof(struct pri));
-		p->fd = fd;
-		p->read_func = rd;
-		p->write_func = wr;
-		p->userdata = userdata;
-		p->localtype = node;
-		p->switchtype = switchtype;
-		p->cref = 1;
-		p->sapi = Q921_SAPI_CALL_CTRL;
-		p->tei = 0;
-		p->nsf = PRI_NSF_NONE;
-		p->protodisc = Q931_PROTOCOL_DISCRIMINATOR;
-		p->master = master;
-		p->callpool = &p->localpool;
-		pri_default_timers(p, switchtype);
-#ifdef LIBPRI_COUNTERS
-		p->q921_rxcount = 0;
-		p->q921_txcount = 0;
-		p->q931_rxcount = 0;
-		p->q931_txcount = 0;
-#endif
-		if (switchtype == PRI_SWITCH_GR303_EOC) {
-			p->protodisc = GR303_PROTOCOL_DISCRIMINATOR;
-			p->sapi = Q921_SAPI_GR303_EOC;
-			p->tei = Q921_TEI_GR303_EOC_OPS;
-			p->subchannel = __pri_new(-1, node, PRI_SWITCH_GR303_EOC_PATH, p, NULL, NULL, NULL);
-			if (!p->subchannel) {
-				free(p);
-				p = NULL;
-			}
-		} else if (switchtype == PRI_SWITCH_GR303_TMC) {
-			p->protodisc = GR303_PROTOCOL_DISCRIMINATOR;
-			p->sapi = Q921_SAPI_GR303_TMC_CALLPROC;
-			p->tei = Q921_TEI_GR303_TMC_CALLPROC;
-			p->subchannel = __pri_new(-1, node, PRI_SWITCH_GR303_TMC_SWITCHING, p, NULL, NULL, NULL);
-			if (!p->subchannel) {
-				free(p);
-				p = NULL;
-			}
-		} else if (switchtype == PRI_SWITCH_GR303_TMC_SWITCHING) {
-			p->protodisc = GR303_PROTOCOL_DISCRIMINATOR;
-			p->sapi = Q921_SAPI_GR303_TMC_SWITCHING;
-			p->tei = Q921_TEI_GR303_TMC_SWITCHING;
-		} else if (switchtype == PRI_SWITCH_GR303_EOC_PATH) {
-			p->protodisc = GR303_PROTOCOL_DISCRIMINATOR;
-			p->sapi = Q921_SAPI_GR303_EOC;
-			p->tei = Q921_TEI_GR303_EOC_PATH;
-		}
-		/* Start Q.921 layer, Wait if we're the network */
-		if (p)
-			q921_start(p, p->localtype == PRI_CPE);
+
+	if (!(p = calloc(1, sizeof(*p))))
+		return NULL;
+
+	p->bri = bri;
+	p->fd = fd;
+	p->read_func = rd;
+	p->write_func = wr;
+	p->userdata = userdata;
+	p->localtype = node;
+	p->switchtype = switchtype;
+	p->cref = 1;
+	p->sapi = (tei == Q921_TEI_GROUP) ? Q921_SAPI_LAYER2_MANAGEMENT : Q921_SAPI_CALL_CTRL;
+	p->tei = tei;
+	p->nsf = PRI_NSF_NONE;
+	p->protodisc = Q931_PROTOCOL_DISCRIMINATOR;
+	p->master = master;
+	p->callpool = &p->localpool;
+	pri_default_timers(p, switchtype);
+	if (master) {
+		pri_set_debug(p, master->debug);
+		if (master->sendfacility)
+			pri_facility_enable(p);
 	}
+#ifdef LIBPRI_COUNTERS
+	p->q921_rxcount = 0;
+	p->q921_txcount = 0;
+	p->q931_rxcount = 0;
+	p->q931_txcount = 0;
+#endif
+	if (switchtype == PRI_SWITCH_GR303_EOC) {
+		p->protodisc = GR303_PROTOCOL_DISCRIMINATOR;
+		p->sapi = Q921_SAPI_GR303_EOC;
+		p->tei = Q921_TEI_GR303_EOC_OPS;
+		p->subchannel = __pri_new_tei(-1, node, PRI_SWITCH_GR303_EOC_PATH, p, NULL, NULL, NULL, Q921_TEI_GR303_EOC_PATH, 0);
+		if (!p->subchannel) {
+			free(p);
+			p = NULL;
+		}
+	} else if (switchtype == PRI_SWITCH_GR303_TMC) {
+		p->protodisc = GR303_PROTOCOL_DISCRIMINATOR;
+		p->sapi = Q921_SAPI_GR303_TMC_CALLPROC;
+		p->tei = Q921_TEI_GR303_TMC_CALLPROC;
+		p->subchannel = __pri_new_tei(-1, node, PRI_SWITCH_GR303_TMC_SWITCHING, p, NULL, NULL, NULL, Q921_TEI_GR303_TMC_SWITCHING, 0);
+		if (!p->subchannel) {
+			free(p);
+			p = NULL;
+		}
+	} else if (switchtype == PRI_SWITCH_GR303_TMC_SWITCHING) {
+		p->protodisc = GR303_PROTOCOL_DISCRIMINATOR;
+		p->sapi = Q921_SAPI_GR303_TMC_SWITCHING;
+		p->tei = Q921_TEI_GR303_TMC_SWITCHING;
+	} else if (switchtype == PRI_SWITCH_GR303_EOC_PATH) {
+		p->protodisc = GR303_PROTOCOL_DISCRIMINATOR;
+		p->sapi = Q921_SAPI_GR303_EOC;
+		p->tei = Q921_TEI_GR303_EOC_PATH;
+	}
+	/* Start Q.921 layer, Wait if we're the network */
+	if (p)
+		q921_start(p, p->localtype == PRI_CPE);
+	
 	return p;
 }
 
@@ -270,7 +288,15 @@ int pri_restart(struct pri *pri)
 
 struct pri *pri_new(int fd, int nodetype, int switchtype)
 {
-	return __pri_new(fd, nodetype, switchtype, NULL, __pri_read, __pri_write, NULL);
+	return __pri_new_tei(fd, nodetype, switchtype, NULL, __pri_read, __pri_write, NULL, Q921_TEI_PRI, 0);
+}
+
+struct pri *pri_new_bri(int fd, int ptpmode, int nodetype, int switchtype)
+{
+	if (ptpmode)
+		return __pri_new_tei(fd, nodetype, switchtype, NULL, __pri_read, __pri_write, NULL, Q921_TEI_PRI, 1);
+	else
+		return __pri_new_tei(fd, nodetype, switchtype, NULL, __pri_read, __pri_write, NULL, Q921_TEI_GROUP, 1);
 }
 
 struct pri *pri_new_cb(int fd, int nodetype, int switchtype, pri_io_cb io_read, pri_io_cb io_write, void *userdata)
@@ -279,7 +305,7 @@ struct pri *pri_new_cb(int fd, int nodetype, int switchtype, pri_io_cb io_read,
 		io_read = __pri_read;
 	if (!io_write)
 		io_write = __pri_write;
-	return __pri_new(fd, nodetype, switchtype, NULL, io_read, io_write, userdata);
+	return __pri_new_tei(fd, nodetype, switchtype, NULL, io_read, io_write, userdata, Q921_TEI_PRI, 0);
 }
 
 void *pri_get_userdata(struct pri *pri)
@@ -556,6 +582,14 @@ int pri_channel_bridge(q931_call *call1, q931_call *call2)
 			else
 				return 0;
 			break;
+		case PRI_SWITCH_QSIG:
+			call1->bridged_call = call2;
+			call2->bridged_call = call1;
+			if (anfpr_initiate_transfer(call1->pri, call1, call2))
+				return -1;
+			else
+				return 0;
+			break;
 		default:
 			return -1;
 	}
@@ -753,6 +787,11 @@ void pri_set_overlapdial(struct pri *pri,int state)
 	pri->overlapdial = state;
 }
 
+void pri_set_inbanddisconnect(struct pri *pri, unsigned int enable)
+{
+	pri->acceptinbanddisconnect = (enable != 0);
+}
+
 int pri_fd(struct pri *pri)
 {
 	return pri->fd;
@@ -821,7 +860,7 @@ void pri_enslave(struct pri *master, struct pri *slave)
 struct pri_sr *pri_sr_new(void)
 {
 	struct pri_sr *req;
-	req = malloc(sizeof(struct pri_sr));
+	req = malloc(sizeof(*req));
 	if (req) 
 		pri_sr_init(req);
 	return req;
