@@ -3,23 +3,28 @@
  *
  * Written by Matthew Fredrickson <creslin@digium.com>
  *
- * Copyright (C) 2004-2005, Digium
+ * Copyright (C) 2004-2005, Digium, Inc.
  * All Rights Reserved.
+ */
+
+/*
+ * See http://www.asterisk.org for more information about
+ * the Asterisk project. Please do not directly contact
+ * any of the maintainers of this project for assistance;
+ * the project provides a web site, mailing lists and IRC
+ * channels for your use.
  *
- * This program is free software; you can redistribute it and/or modify
- * it under the terms of the GNU General Public License as published by
- * the Free Software Foundation; either version 2 of the License, or
- * (at your option) any later version.
- * 
- * This program is distributed in the hope that it will be useful,
- * but WITHOUT ANY WARRANTY; without even the implied warranty of
- * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
- * GNU General Public License for more details.
- * 
- * You should have received a copy of the GNU General Public License
- * along with this program; if not, write to the Free Software
- * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA. 
+ * This program is free software, distributed under the terms of
+ * the GNU General Public License Version 2 as published by the
+ * Free Software Foundation. See the LICENSE file included with
+ * this program for more details.
  *
+ * In addition, when this program is distributed with Asterisk in
+ * any form that would qualify as a 'combined work' or as a
+ * 'derivative work' (but not mere aggregation), you can redistribute
+ * and/or modify the combination under the terms of the license
+ * provided with that copy of Asterisk, instead of the license
+ * terms granted here.
  */
 
 #include "compat.h"
@@ -167,9 +172,18 @@ static unsigned char get_invokeid(struct pri *pri)
 struct addressingdataelements_presentednumberunscreened {
 	char partyaddress[21];
 	char partysubaddress[21];
-	int  npi;
-	int  ton;
-	int  pres;
+	int  npi;       /* Numbering Plan Indicator */
+	int  ton;       /* Type Of Number */
+	int  pres;      /* Presentation */
+};
+
+struct addressingdataelements_presentednumberscreened {
+	char partyaddress[21];
+	char partysubaddress[21];
+	int  npi;       /* Numbering Plan Indicator */
+	int  ton;       /* Type Of Number */
+	int  pres;      /* Presentation */
+	int  scrind;    /* Screening Indicator */
 };
 
 #define PRI_CHECKOVERFLOW(size) \
@@ -453,6 +467,35 @@ static int rose_public_party_number_decode(struct pri *pri, q931_call *call, uns
 	return -1;
 }
 
+static int rose_private_party_number_decode(struct pri *pri, q931_call *call, unsigned char *data, int len, struct addressingdataelements_presentednumberunscreened *value)
+{
+	int i = 0;
+	struct rose_component *comp = NULL;
+	unsigned char *vdata = data;
+	int ton;
+	int res = 0;
+
+	if (len < 2)
+	return -1;
+
+	do {
+		GET_COMPONENT(comp, i, vdata, len);
+		CHECK_COMPONENT(comp, ASN1_ENUMERATED, "Don't know what to do with PrivatePartyNumber ROSE component type 0x%x\n");
+		ASN1_GET_INTEGER(comp, ton);
+		NEXT_COMPONENT(comp, i);
+		ton = typeofnumber_for_q931(pri, ton);
+
+		res = rose_number_digits_decode(pri, call, &vdata[i], len-i, value);
+		if (res < 0)
+		  return -1;
+		value->ton = ton;
+
+		return res + 3;
+
+	} while(0);
+	return -1;
+}
+
 static int rose_address_decode(struct pri *pri, q931_call *call, unsigned char *data, int len, struct addressingdataelements_presentednumberunscreened *value)
 {
 	int i = 0;
@@ -503,9 +546,11 @@ static int rose_address_decode(struct pri *pri, q931_call *call, unsigned char *
 			pri_message(pri, "!! telexPartyNumber isn't handled\n");
 			return -1;
 		case (ASN1_CONTEXT_SPECIFIC | ASN1_CONSTRUCTOR | ASN1_TAG_5):	/* [5] priavePartyNumber */
-			pri_message(pri, "!! privatePartyNumber isn't handled\n");
-			value->npi = PRI_NPI_PRIVATE;
+			res = rose_private_party_number_decode(pri, call, comp->data, comp->len, value);
+			if (res < 0)
 			return -1;
+			value->npi = PRI_NPI_PRIVATE;
+			break;
 		case (ASN1_CONTEXT_SPECIFIC | ASN1_CONSTRUCTOR | ASN1_TAG_8):	/* [8] nationalStandardPartyNumber */
 			res = rose_number_digits_decode(pri, call, comp->data, comp->len, value);
 			if (res < 0)
@@ -589,6 +634,8 @@ static int rose_diverting_leg_information2_decode(struct pri *pri, q931_call *ca
 	struct rose_component *comp = NULL;
 	unsigned char *vdata = sequence->data;
 	int res = 0;
+	memset(&divertingnr, 0, sizeof(divertingnr));
+	memset(&originalcallednr, 0, sizeof(originalcallednr));
 
 	/* Data checks */
 	if (sequence->type != (ASN1_CONSTRUCTOR | ASN1_SEQUENCE)) { /* Constructed Sequence */
@@ -670,6 +717,9 @@ static int rose_diverting_leg_information2_decode(struct pri *pri, q931_call *ca
 				if (pri->debug & PRI_DEBUG_APDU)
 					pri_message(pri, "    Received Originally Called Name '%s'\n", origcalledname);
 				break;
+			case (ASN1_CONTEXT_SPECIFIC | ASN1_CONSTRUCTOR | ASN1_TAG_5):
+				pri_message(pri, "!! Ignoring DivertingLegInformation2 component 0x%X\n", comp->type);
+				break;
 			default:
 				if (comp->type == 0 && comp->len == 0) {
 					break; /* Found termination characters */
@@ -684,11 +734,13 @@ static int rose_diverting_leg_information2_decode(struct pri *pri, q931_call *ca
 			call->redirectingpres = divertingnr.pres;
 			call->redirectingreason = diversion_reason;
 			libpri_copy_string(call->redirectingnum, divertingnr.partyaddress, sizeof(call->redirectingnum));
+			pri_message(pri, "    Received redirectingnum '%s' (%d)\n", call->redirectingnum, (int)call->redirectingnum[0]);
 		}
 		if (originalcallednr.pres >= 0) {
 			call->origcalledplan = originalcallednr.npi;
 			call->origcalledpres = originalcallednr.pres;
 			libpri_copy_string(call->origcallednum, originalcallednr.partyaddress, sizeof(call->origcallednum));
+			pri_message(pri, "    Received origcallednum '%s' (%d)\n", call->origcallednum, (int)call->origcallednum[0]);
 		}
 		libpri_copy_string(call->redirectingname, redirectingname, sizeof(call->redirectingname));
 		libpri_copy_string(call->origcalledname, origcalledname, sizeof(call->origcalledname));
@@ -1134,6 +1186,162 @@ int eect_initiate_transfer(struct pri *pri, q931_call *c1, q931_call *c2)
 }
 /* End EECT */
 
+static int anfpr_pathreplacement_respond(struct pri *pri, q931_call *call, q931_ie *ie)
+{
+	int res;
+	
+	res = pri_call_apdu_queue_cleanup(call->bridged_call);
+	if (res) {
+	        pri_message(pri, "Could not Clear queue ADPU\n");
+	        return -1;
+	}
+	
+	/* Send message */
+	res = pri_call_apdu_queue(call->bridged_call, Q931_FACILITY, ie->data, ie->len, NULL, NULL);
+	if (res) {
+	        pri_message(pri, "Could not queue ADPU in facility message\n");
+	        return -1;
+	}
+	
+	/* Remember that if we queue a facility IE for a facility message we
+	 * have to explicitly send the facility message ourselves */
+	
+	res = q931_facility(call->bridged_call->pri, call->bridged_call);
+	if (res) {
+		pri_message(pri, "Could not schedule facility message for call %d\n", call->bridged_call->cr);
+		return -1;
+	}
+
+	return 0;
+}
+/* AFN-PR */
+extern int anfpr_initiate_transfer(struct pri *pri, q931_call *c1, q931_call *c2)
+{
+	/* Did all the tests to see if we're on the same PRI and
+	 * are on a compatible switchtype */
+	/* TODO */
+	int i = 0;
+	int res = 0;
+	unsigned char buffer[255] = "";
+	unsigned short call_reference = c2->cr;
+	struct rose_component *comp = NULL, *compstk[10];
+	unsigned char buffer2[255] = "";
+	int compsp = 0;
+	static unsigned char op_tag[] = {
+		0x0C,
+	};
+	
+	/* Channel 1 */
+	buffer[i++] = (ASN1_CONTEXT_SPECIFIC | Q932_PROTOCOL_EXTENSIONS);
+	/* Interpretation component */
+	
+	ASN1_ADD_SIMPLE(comp, COMP_TYPE_NFE, buffer, i);
+	ASN1_PUSH(compstk, compsp, comp);
+	ASN1_ADD_BYTECOMP(comp, (ASN1_CONTEXT_SPECIFIC | ASN1_TAG_0), buffer, i, 0);
+	ASN1_ADD_BYTECOMP(comp, (ASN1_CONTEXT_SPECIFIC | ASN1_TAG_2), buffer, i, 0);
+	ASN1_FIXUP(compstk, compsp, buffer, i);
+	
+	ASN1_ADD_BYTECOMP(comp, COMP_TYPE_INTERPRETATION, buffer, i, 2);    /* reject - to get feedback from QSIG switch */
+	
+	ASN1_ADD_SIMPLE(comp, COMP_TYPE_INVOKE, buffer, i);
+	ASN1_PUSH(compstk, compsp, comp);
+	
+	ASN1_ADD_BYTECOMP(comp, ASN1_INTEGER, buffer, i, get_invokeid(pri));
+	
+	res = asn1_string_encode(ASN1_INTEGER, &buffer[i], sizeof(buffer)-i, sizeof(op_tag), op_tag, sizeof(op_tag));
+	if (res < 0)
+		return -1;
+	i += res;
+	
+	ASN1_ADD_SIMPLE(comp, (ASN1_SEQUENCE | ASN1_CONSTRUCTOR), buffer, i);
+	ASN1_PUSH(compstk, compsp, comp);
+	buffer[i++] = (0x0a);
+	buffer[i++] = (0x01);
+	buffer[i++] = (0x00);
+	buffer[i++] = (0x81);
+	buffer[i++] = (0x00);
+	buffer[i++] = (0x0a);
+	buffer[i++] = (0x01);
+	buffer[i++] = (0x01);
+	ASN1_ADD_WORDCOMP(comp, ASN1_INTEGER, buffer, i, call_reference);
+	ASN1_FIXUP(compstk, compsp, buffer, i);
+	ASN1_FIXUP(compstk, compsp, buffer, i);
+	
+	res = pri_call_apdu_queue(c1, Q931_FACILITY, buffer, i, NULL, NULL);
+	if (res) {
+		pri_message(pri, "Could not queue ADPU in facility message\n");
+		return -1;
+	}
+	
+	/* Remember that if we queue a facility IE for a facility message we
+	 * have to explicitly send the facility message ourselves */
+	
+	res = q931_facility(c1->pri, c1);
+	if (res) {
+		pri_message(pri, "Could not schedule facility message for call %d\n", c1->cr);
+		return -1;
+	}
+	
+	/* Channel 2 */
+	i = 0;
+	res = 0;
+	compsp = 0;
+	
+	buffer2[i++] = (ASN1_CONTEXT_SPECIFIC | Q932_PROTOCOL_EXTENSIONS);
+	/* Interpretation component */
+	
+	ASN1_ADD_SIMPLE(comp, COMP_TYPE_NFE, buffer2, i);
+	ASN1_PUSH(compstk, compsp, comp);
+	ASN1_ADD_BYTECOMP(comp, (ASN1_CONTEXT_SPECIFIC | ASN1_TAG_0), buffer2, i, 0);
+	ASN1_ADD_BYTECOMP(comp, (ASN1_CONTEXT_SPECIFIC | ASN1_TAG_2), buffer2, i, 0);
+	ASN1_FIXUP(compstk, compsp, buffer2, i);
+	
+	ASN1_ADD_BYTECOMP(comp, COMP_TYPE_INTERPRETATION, buffer2, i, 2);  /* reject */
+	
+	ASN1_ADD_SIMPLE(comp, COMP_TYPE_INVOKE, buffer2, i);
+	ASN1_PUSH(compstk, compsp, comp);
+	
+	ASN1_ADD_BYTECOMP(comp, ASN1_INTEGER, buffer2, i, get_invokeid(pri));
+	
+	res = asn1_string_encode(ASN1_INTEGER, &buffer2[i], sizeof(buffer2)-i, sizeof(op_tag), op_tag, sizeof(op_tag));
+	if (res < 0)
+		return -1;
+	i += res;
+	
+	ASN1_ADD_SIMPLE(comp, (ASN1_SEQUENCE | ASN1_CONSTRUCTOR), buffer2, i);
+	ASN1_PUSH(compstk, compsp, comp);
+	buffer2[i++] = (0x0a);
+	buffer2[i++] = (0x01);
+	buffer2[i++] = (0x01);
+	buffer2[i++] = (0x81);
+	buffer2[i++] = (0x00);
+	buffer2[i++] = (0x0a);
+	buffer2[i++] = (0x01);
+	buffer2[i++] = (0x01);
+	ASN1_ADD_WORDCOMP(comp, ASN1_INTEGER, buffer2, i, call_reference);
+	ASN1_FIXUP(compstk, compsp, buffer2, i);
+	ASN1_FIXUP(compstk, compsp, buffer2, i);
+	
+	
+	res = pri_call_apdu_queue(c2, Q931_FACILITY, buffer2, i, NULL, NULL);
+	if (res) {
+		pri_message(pri, "Could not queue ADPU in facility message\n");
+		return -1;
+	}
+	
+	/* Remember that if we queue a facility IE for a facility message we
+	 * have to explicitly send the facility message ourselves */
+	
+	res = q931_facility(c2->pri, c2);
+	if (res) {
+		pri_message(pri, "Could not schedule facility message for call %d\n", c1->cr);
+		return -1;
+	}
+	
+	return 0;
+}
+/* End AFN-PR */
+
 /* AOC */
 static int aoc_aoce_charging_request_decode(struct pri *pri, q931_call *call, unsigned char *data, int len) 
 {
@@ -1340,7 +1548,433 @@ static int aoc_aoce_charging_unit_encode(struct pri *pri, q931_call *c, long cha
 }
 /* End AOC */
 
-int rose_reject_decode(struct pri *pri, q931_call *call, unsigned char *data, int len)
+/* ===== Call Transfer Supplementary Service (ECMA-178) ===== */
+
+static int rose_party_number_decode(struct pri *pri, q931_call *call, unsigned char *data, int len, struct addressingdataelements_presentednumberunscreened *value)
+{
+	int i = 0;
+	int size = 0;
+	struct rose_component *comp = NULL;
+	unsigned char *vdata = data;
+
+
+	do {
+		GET_COMPONENT(comp, i, vdata, len);
+
+		switch(comp->type) {
+			case (ASN1_CONTEXT_SPECIFIC | ASN1_CONSTRUCTOR | ASN1_TAG_0):   /* [0] IMPLICIT NumberDigits -- default: unknownPartyNumber */
+				if (pri->debug & PRI_DEBUG_APDU)
+					pri_message(pri, "     PartyNumber: UnknownPartyNumber len=%d\n", len);
+				size = rose_number_digits_decode(pri, call, comp->data, comp->len, value);
+				if (size < 0)
+					return -1;
+				value->npi = PRI_NPI_UNKNOWN;
+				value->ton = PRI_TON_UNKNOWN;
+				break;
+
+			case (ASN1_CONTEXT_SPECIFIC | ASN1_CONSTRUCTOR | ASN1_TAG_1):   /* [1] IMPLICIT PublicPartyNumber */
+				if (pri->debug & PRI_DEBUG_APDU)
+					pri_message(pri, "     PartyNumber: PublicPartyNumber len=%d\n", len);
+				size = rose_public_party_number_decode(pri, call, comp->data, comp->len, value);
+				if (size < 0)
+					return -1;
+				value->npi = PRI_NPI_E163_E164;
+				break;
+
+			case (ASN1_CONTEXT_SPECIFIC | ASN1_CONSTRUCTOR | ASN1_TAG_3):   /* [3] IMPLICIT NumberDigits -- not used: dataPartyNumber */
+				pri_message(pri, "!! PartyNumber: dataPartyNumber is reserved!\n");
+				size = rose_number_digits_decode(pri, call, comp->data, comp->len, value);
+				if (size < 0)
+					return -1;
+				value->npi = PRI_NPI_X121 /* ??? */;
+				value->ton = PRI_TON_UNKNOWN /* ??? */;
+				break;
+
+			case (ASN1_CONTEXT_SPECIFIC | ASN1_CONSTRUCTOR | ASN1_TAG_4):   /* [4] IMPLICIT NumberDigits -- not used: telexPartyNumber */
+				pri_message(pri, "!! PartyNumber: telexPartyNumber is reserved!\n");
+				size = rose_number_digits_decode(pri, call, comp->data, comp->len, value);
+				if (size < 0)
+					return -1;
+				value->npi = PRI_NPI_F69 /* ??? */;
+				value->ton = PRI_TON_UNKNOWN /* ??? */;
+				break;
+
+			case (ASN1_CONTEXT_SPECIFIC | ASN1_CONSTRUCTOR | ASN1_TAG_5):   /* [5] IMPLICIT PrivatePartyNumber */
+				if (pri->debug & PRI_DEBUG_APDU)
+					pri_message(pri, "     PartyNumber: PrivatePartyNumber len=%d\n", len);
+				size = rose_private_party_number_decode(pri, call, comp->data, comp->len, value);
+				if (size < 0)
+					return -1;
+ 				value->npi = PRI_NPI_PRIVATE;
+				break;
+
+			case (ASN1_CONTEXT_SPECIFIC | ASN1_CONSTRUCTOR | ASN1_TAG_8):   /* [8] IMPLICIT NumberDigits -- not used: nationalStandatdPartyNumber */
+				pri_message(pri, "!! PartyNumber: nationalStandardPartyNumber is reserved!\n");
+				size = rose_number_digits_decode(pri, call, comp->data, comp->len, value);
+				if (size < 0)
+					return -1;
+				value->npi = PRI_NPI_NATIONAL;
+				value->ton = PRI_TON_NATIONAL;
+				break;
+
+			default:
+				pri_message(pri, "Invalid PartyNumber component 0x%X\n", comp->type);
+				return -1;
+		}
+		ASN1_FIXUP_LEN(comp, size);
+		if (pri->debug & PRI_DEBUG_APDU)
+			pri_message(pri, "     PartyNumber: '%s' size=%d len=%d\n", value->partyaddress, size, len);
+		return size;
+	}
+	while (0);
+
+	return -1;
+}
+
+
+static int rose_number_screened_decode(struct pri *pri, q931_call *call, unsigned char *data, int len, struct addressingdataelements_presentednumberscreened *value)
+{
+	int i = 0;
+	int size = 0;
+	struct rose_component *comp = NULL;
+	unsigned char *vdata = data;
+
+	int scrind = -1;
+
+	do {
+		/* Party Number */
+		GET_COMPONENT(comp, i, vdata, len);
+		size = rose_party_number_decode(pri, call, (u_int8_t *)comp, comp->len + 2, (struct addressingdataelements_presentednumberunscreened*) value);
+		if (size < 0)
+			return -1;
+		comp->len = size;
+		NEXT_COMPONENT(comp, i);
+
+		/* Screening Indicator */
+		GET_COMPONENT(comp, i, vdata, len);
+		CHECK_COMPONENT(comp, ASN1_ENUMERATED, "Don't know what to do with NumberScreened ROSE component type 0x%x\n");
+		ASN1_GET_INTEGER(comp, scrind);
+		// Todo: scrind = screeningindicator_for_q931(pri, scrind);
+		NEXT_COMPONENT(comp, i);
+
+		value->scrind = scrind;
+
+		if (pri->debug & PRI_DEBUG_APDU)
+			pri_message(pri, "     NumberScreened: '%s' ScreeningIndicator=%d  i=%d  len=%d\n", value->partyaddress, scrind, i, len);
+
+		return i-2;  // We do not have a sequence header here.
+	}
+	while (0);
+
+	return -1;
+}
+
+
+static int rose_presented_number_screened_decode(struct pri *pri, q931_call *call, unsigned char *data, int len, struct addressingdataelements_presentednumberscreened *value)
+{
+	int i = 0;
+	int size = 0;
+	struct rose_component *comp = NULL;
+	unsigned char *vdata = data;
+
+	/* Fill in default values */
+	value->ton = PRI_TON_UNKNOWN;
+	value->npi = PRI_NPI_UNKNOWN;
+	value->pres = -1; /* Data is not available */
+
+	do {
+		GET_COMPONENT(comp, i, vdata, len);
+
+		switch(comp->type) {
+			case (ASN1_CONTEXT_SPECIFIC | ASN1_CONSTRUCTOR | ASN1_TAG_0):   /* [0] IMPLICIT presentationAllowedNumber */
+				if (pri->debug & PRI_DEBUG_APDU)
+					pri_message(pri, "     PresentedNumberScreened: presentationAllowedNumber comp->len=%d\n", comp->len);
+				value->pres = PRES_ALLOWED_USER_NUMBER_PASSED_SCREEN;
+				size = rose_number_screened_decode(pri, call, comp->data, comp->len, value);
+				if (size < 0)
+					return -1;
+				ASN1_FIXUP_LEN(comp, size);
+				return size + 2;
+
+			case (ASN1_CONTEXT_SPECIFIC | ASN1_TAG_1):    /* [1] IMPLICIT presentationRestricted */
+				if (pri->debug & PRI_DEBUG_APDU)
+					pri_message(pri, "     PresentedNumberScreened: presentationRestricted comp->len=%d\n", comp->len);
+				if (comp->len != 0) { /* must be NULL */
+					pri_error(pri, "!! Invalid PresentationRestricted component received (len != 0)\n");
+					return -1;
+				}
+				value->pres = PRES_PROHIB_USER_NUMBER_PASSED_SCREEN;
+				return 2;
+
+			case (ASN1_CONTEXT_SPECIFIC | ASN1_TAG_2):    /* [2] IMPLICIT numberNotAvailableDueToInterworking */
+				if (pri->debug & PRI_DEBUG_APDU)
+					pri_message(pri, "     PresentedNumberScreened: NumberNotAvailableDueToInterworking comp->len=%d\n", comp->len);
+				if (comp->len != 0) { /* must be NULL */
+					pri_error(pri, "!! Invalid NumberNotAvailableDueToInterworking component received (len != 0)\n");
+					return -1;
+				}
+				value->pres = PRES_NUMBER_NOT_AVAILABLE;
+				if (pri->debug & PRI_DEBUG_APDU)
+					pri_message(pri, "     PresentedNumberScreened: numberNotAvailableDueToInterworking Type=0x%X  i=%d len=%d size=%d\n", comp->type, i, len);
+				return 2;
+
+			case (ASN1_CONTEXT_SPECIFIC | ASN1_CONSTRUCTOR | ASN1_TAG_3):    /* [3] IMPLICIT presentationRestrictedNumber */
+				if (pri->debug & PRI_DEBUG_APDU)
+					pri_message(pri, "     PresentedNumberScreened: presentationRestrictedNumber comp->len=%d\n", comp->len);
+				value->pres = PRES_PROHIB_USER_NUMBER_PASSED_SCREEN;
+				size = rose_number_screened_decode(pri, call, comp->data, comp->len, value);
+				if (size < 0)
+					return -1;
+				ASN1_FIXUP_LEN(comp, size);
+				return size + 2;
+
+			default:
+				pri_message(pri, "Invalid PresentedNumberScreened component 0x%X\n", comp->type);
+		}
+		return -1;
+	}
+	while (0);
+
+	return -1;
+}
+
+
+static int rose_call_transfer_complete_decode(struct pri *pri, q931_call *call, struct rose_component *sequence, int len)
+{
+	int i = 0;
+	struct rose_component *comp = NULL;
+	unsigned char *vdata = sequence->data;
+	int res = 0;
+
+	int end_designation = 0;
+	struct addressingdataelements_presentednumberscreened redirection_number;
+	char redirection_name[50] = "";
+	int call_status = 0;
+	redirection_number.partyaddress[0] = 0;
+	redirection_number.partysubaddress[0] = 0;
+	call->callername[0] = 0;
+	call->callernum[0] = 0;
+
+
+	/* Data checks */
+	if (sequence->type != (ASN1_CONSTRUCTOR | ASN1_SEQUENCE)) { /* Constructed Sequence */
+		pri_message(pri, "Invalid callTransferComplete argument. (Not a sequence)\n");
+		return -1;
+	}
+
+	if (sequence->len == ASN1_LEN_INDEF) {
+		len -= 4; /* For the 2 extra characters at the end
+					   * and two characters of header */
+	} else
+		len -= 2;
+
+	if (pri->debug & PRI_DEBUG_APDU)
+		pri_message(pri, "     CT-Complete: len=%d\n", len);
+
+	do {
+		/* End Designation */
+		GET_COMPONENT(comp, i, vdata, len);
+		CHECK_COMPONENT(comp, ASN1_ENUMERATED, "Invalid endDesignation type 0x%X of ROSE callTransferComplete component received\n");
+		ASN1_GET_INTEGER(comp, end_designation);
+		NEXT_COMPONENT(comp, i);
+		if (pri->debug & PRI_DEBUG_APDU)
+			pri_message(pri, "     CT-Complete: Received endDesignation=%d\n", end_designation);
+
+
+		/* Redirection Number */
+		GET_COMPONENT(comp, i, vdata, len);
+		res = rose_presented_number_screened_decode(pri, call, (u_int8_t *)comp, comp->len + 2, &redirection_number);
+		if (res < 0)
+			return -1;
+		comp->len = res;
+		if (res > 2) {
+			if (pri->debug & PRI_DEBUG_APDU)
+				pri_message(pri, "     CT-Complete: Received redirectionNumber=%s\n", redirection_number.partyaddress);
+			strncpy(call->callernum, redirection_number.partyaddress, 20);
+			call->callernum[20] = 0;
+		}
+		NEXT_COMPONENT(comp, i);
+
+
+#if 0 /* This one is optional. How do we check if it is there? */
+		/* Basic Call Info Elements */
+		GET_COMPONENT(comp, i, vdata, len);
+		NEXT_COMPONENT(comp, i);
+#endif
+
+
+		/* Redirection Name */
+		GET_COMPONENT(comp, i, vdata, len);
+		res = asn1_name_decode((u_int8_t *)comp, comp->len + 2, redirection_name, sizeof(redirection_name));
+		if (res < 0)
+			return -1;
+		memcpy(call->callername, comp->data, comp->len);
+		call->callername[comp->len] = 0;
+		ASN1_FIXUP_LEN(comp, res);
+		comp->len = res;
+		NEXT_COMPONENT(comp, i);
+		if (pri->debug & PRI_DEBUG_APDU)
+			pri_message(pri, "     CT-Complete: Received redirectionName '%s'\n", redirection_name);
+
+
+		/* Call Status */
+		GET_COMPONENT(comp, i, vdata, len);
+		CHECK_COMPONENT(comp, ASN1_ENUMERATED, "Invalid callStatus type 0x%X of ROSE callTransferComplete component received\n");
+		ASN1_GET_INTEGER(comp, call_status);
+		NEXT_COMPONENT(comp, i);
+		if (pri->debug & PRI_DEBUG_APDU)
+			pri_message(pri, "     CT-Complete: Received callStatus=%d\n", call_status);
+
+
+		/* Argument Extension */
+#if 0 /* Not supported */
+		GET_COMPONENT(comp, i, vdata, len);
+		switch (comp->type) {
+			case (ASN1_CONTEXT_SPECIFIC | ASN1_TAG_9):   /* [9] IMPLICIT Extension */
+				res = rose_extension_decode(pri, call, comp->data, comp->len, &redirection_number);
+				if (res < 0)
+					return -1;
+				ASN1_FIXUP_LEN(comp, res);
+				comp->len = res;
+
+			case (ASN1_CONTEXT_SPECIFIC | ASN1_TAG_10):    /* [10] IMPLICIT SEQUENCE OF Extension */
+				res = rose_sequence_of_extension_decode(pri, call, comp->data, comp->len, &redirection_number);
+				if (res < 0)
+					return -1;
+				ASN1_FIXUP_LEN(comp, res);
+				comp->len = res;
+
+			default:
+			pri_message(pri, "     CT-Complete: !! Unknown argumentExtension received 0x%X\n", comp->type);
+			return -1;
+		}
+#else
+		GET_COMPONENT(comp, i, vdata, len);
+		ASN1_FIXUP_LEN(comp, res);
+		NEXT_COMPONENT(comp, i);
+#endif
+
+		if(i < len)
+			pri_message(pri, "     CT-Complete: !! not all information is handled !! i=%d / len=%d\n", i, len);
+
+		return 0;
+	}
+	while (0);
+
+	return -1;
+}
+
+
+static int rose_call_transfer_update_decode(struct pri *pri, q931_call *call, struct rose_component *sequence, int len)
+{
+	int i = 0;
+	struct rose_component *comp = NULL;
+	unsigned char *vdata = sequence->data;
+	int res = 0;
+
+	struct addressingdataelements_presentednumberscreened redirection_number;
+	redirection_number.partyaddress[0] = 0;
+	redirection_number.partysubaddress[0] = 0;
+	char redirection_name[50] = "";
+	call->callername[0] = 0;
+	call->callernum[0] = 0;
+
+
+	/* Data checks */
+	if (sequence->type != (ASN1_CONSTRUCTOR | ASN1_SEQUENCE)) { /* Constructed Sequence */
+		pri_message(pri, "Invalid callTransferComplete argument. (Not a sequence)\n");
+		return -1;
+	}
+
+	if (sequence->len == ASN1_LEN_INDEF) {
+		len -= 4; /* For the 2 extra characters at the end
+					   * and two characters of header */
+	} else
+		len -= 2;
+
+	if (pri->debug & PRI_DEBUG_APDU)
+		pri_message(pri, "     CT-Complete: len=%d\n", len);
+
+	do {
+		/* Redirection Number */
+		GET_COMPONENT(comp, i, vdata, len);
+		res = rose_presented_number_screened_decode(pri, call, (u_int8_t *)comp, comp->len + 2, &redirection_number);
+		if (res < 0)
+			return -1;
+		comp->len = res;
+		if (res > 2) {
+			if (pri->debug & PRI_DEBUG_APDU)
+				pri_message(pri, "     CT-Complete: Received redirectionNumber=%s\n", redirection_number.partyaddress);
+			strncpy(call->callernum, redirection_number.partyaddress, 20);
+			call->callernum[20] = 0;
+		}
+		NEXT_COMPONENT(comp, i);
+
+		/* Redirection Name */
+		GET_COMPONENT(comp, i, vdata, len);
+		res = asn1_name_decode((u_int8_t *)comp, comp->len + 2, redirection_name, sizeof(redirection_name));
+		if (res < 0)
+			return -1;
+		memcpy(call->callername, comp->data, comp->len);
+		call->callername[comp->len] = 0;
+		ASN1_FIXUP_LEN(comp, res);
+		comp->len = res;
+		NEXT_COMPONENT(comp, i);
+		if (pri->debug & PRI_DEBUG_APDU)
+			pri_message(pri, "     CT-Complete: Received redirectionName '%s'\n", redirection_name);
+
+
+#if 0 /* This one is optional. How do we check if it is there? */
+		/* Basic Call Info Elements */
+		GET_COMPONENT(comp, i, vdata, len);
+		NEXT_COMPONENT(comp, i);
+#endif
+
+
+		/* Argument Extension */
+#if 0 /* Not supported */
+		GET_COMPONENT(comp, i, vdata, len);
+		switch (comp->type) {
+			case (ASN1_CONTEXT_SPECIFIC | ASN1_TAG_9):   /* [9] IMPLICIT Extension */
+				res = rose_extension_decode(pri, call, comp->data, comp->len, &redirection_number);
+				if (res < 0)
+					return -1;
+				ASN1_FIXUP_LEN(comp, res);
+				comp->len = res;
+
+			case (ASN1_CONTEXT_SPECIFIC | ASN1_TAG_10):    /* [10] IMPLICIT SEQUENCE OF Extension */
+				res = rose_sequence_of_extension_decode(pri, call, comp->data, comp->len, &redirection_number);
+				if (res < 0)
+					return -1;
+				ASN1_FIXUP_LEN(comp, res);
+				comp->len = res;
+
+			default:
+				pri_message(pri, "     CT-Complete: !! Unknown argumentExtension received 0x%X\n", comp->type);
+				return -1;
+		}
+#else
+		GET_COMPONENT(comp, i, vdata, len);
+		ASN1_FIXUP_LEN(comp, res);
+		NEXT_COMPONENT(comp, i);
+#endif
+
+		if(i < len)
+			pri_message(pri, "     CT-Complete: !! not all information is handled !! i=%d / len=%d\n", i, len);
+
+		return 0;
+	}
+	while (0);
+
+	return -1;
+}
+
+
+/* ===== End Call Transfer Supplementary Service (ECMA-178) ===== */
+
+
+
+int rose_reject_decode(struct pri *pri, q931_call *call, q931_ie *ie, unsigned char *data, int len)
 {
 	int i = 0;
 	int problemtag = -1;
@@ -1408,7 +2042,7 @@ int rose_reject_decode(struct pri *pri, q931_call *call, unsigned char *data, in
 	
 	return -1;
 }
-int rose_return_error_decode(struct pri *pri, q931_call *call, unsigned char *data, int len)
+int rose_return_error_decode(struct pri *pri, q931_call *call, q931_ie *ie, unsigned char *data, int len)
 {
 	int i = 0;
 	int errorvalue = -1;
@@ -1472,7 +2106,7 @@ int rose_return_error_decode(struct pri *pri, q931_call *call, unsigned char *da
 	return -1;
 }
 
-int rose_return_result_decode(struct pri *pri, q931_call *call, unsigned char *data, int len)
+int rose_return_result_decode(struct pri *pri, q931_call *call, q931_ie *ie, unsigned char *data, int len)
 {
 	int i = 0;
 	int operationidvalue = -1;
@@ -1534,9 +2168,10 @@ int rose_return_result_decode(struct pri *pri, q931_call *call, unsigned char *d
 	return -1;
 }
 
-int rose_invoke_decode(struct pri *pri, q931_call *call, unsigned char *data, int len)
+int rose_invoke_decode(struct pri *pri, q931_call *call, q931_ie *ie, unsigned char *data, int len)
 {
 	int i = 0;
+	int res = 0;
 	int operation_tag;
 	unsigned char *vdata = data;
 	struct rose_component *comp = NULL, *invokeid = NULL, *operationid = NULL;
@@ -1587,6 +2222,50 @@ int rose_invoke_decode(struct pri *pri, q931_call *call, unsigned char *data, in
 					return -1;
 			}
 			break;
+		case ROSE_CALL_TRANSFER_IDENTIFY:
+			if (pri->debug & PRI_DEBUG_APDU)
+				pri_message(pri, "ROSE %i:   CallTransferIdentify - not handled!\n", operation_tag);
+			dump_apdu (pri, (u_int8_t *)comp, comp->len + 2);
+			return -1;
+		case ROSE_CALL_TRANSFER_ABANDON:
+			if (pri->debug & PRI_DEBUG_APDU)
+				pri_message(pri, "ROSE %i:   CallTransferAbandon - not handled!\n", operation_tag);
+			dump_apdu (pri, (u_int8_t *)comp, comp->len + 2);
+			return -1;
+		case ROSE_CALL_TRANSFER_INITIATE:
+			if (pri->debug & PRI_DEBUG_APDU)
+				pri_message(pri, "ROSE %i:   CallTransferInitiate - not handled!\n", operation_tag);
+			dump_apdu (pri, (u_int8_t *)comp, comp->len + 2);
+			return -1;
+		case ROSE_CALL_TRANSFER_SETUP:
+			if (pri->debug & PRI_DEBUG_APDU)
+				pri_message(pri, "ROSE %i:   CallTransferSetup - not handled!\n", operation_tag);
+			dump_apdu (pri, (u_int8_t *)comp, comp->len + 2);
+			return -1;
+		case ROSE_CALL_TRANSFER_ACTIVE:
+			if (pri->debug & PRI_DEBUG_APDU)
+				pri_message(pri, "ROSE %i:   CallTransferActive - not handled!\n", operation_tag);
+			dump_apdu (pri, (u_int8_t *)comp, comp->len + 2);
+			return -1;
+		case ROSE_CALL_TRANSFER_COMPLETE:
+			if (pri->debug & PRI_DEBUG_APDU)
+			{
+				pri_message(pri, "ROSE %i:   Handle CallTransferComplete\n", operation_tag);
+				dump_apdu (pri, (u_int8_t *)comp, comp->len + 2);
+			}
+			return rose_call_transfer_complete_decode(pri, call, comp, len-i);
+		case ROSE_CALL_TRANSFER_UPDATE:
+			if (pri->debug & PRI_DEBUG_APDU)
+			{
+				pri_message(pri, "ROSE %i:    Handle CallTransferUpdate\n", operation_tag);
+				dump_apdu (pri, (u_int8_t *)comp, comp->len + 2);
+			}
+			return rose_call_transfer_update_decode(pri, call, comp, len-i);
+		case ROSE_SUBADDRESS_TRANSFER:
+			if (pri->debug & PRI_DEBUG_APDU)
+				pri_message(pri, "ROSE %i:   SubaddressTransfer - not handled!\n", operation_tag);
+			dump_apdu (pri, (u_int8_t *)comp, comp->len + 2);
+			return -1;
 		case ROSE_DIVERTING_LEG_INFORMATION2:
 			if (pri->debug & PRI_DEBUG_APDU)
 				pri_message(pri, "  Handle DivertingLegInformation2\n");
@@ -1641,6 +2320,15 @@ int rose_invoke_decode(struct pri *pri, q931_call *call, unsigned char *data, in
 				dump_apdu (pri, (u_int8_t *)comp, comp->len + 2);
 			}
 			return -1;
+                case SS_ANFPR_PATHREPLACEMENT:
+			/* Clear Queue */
+			res = pri_call_apdu_queue_cleanup(call->bridged_call);
+			if (res) {
+			        pri_message(pri, "Could not Clear queue ADPU\n");
+			        return -1;
+			}
+			anfpr_pathreplacement_respond(pri, call, ie);
+                        break;
 		default:
 			if (pri->debug & PRI_DEBUG_APDU) {
 				pri_message(pri, "!! Unable to handle ROSE operation %d", operation_tag);
@@ -1661,19 +2349,16 @@ int pri_call_apdu_queue(q931_call *call, int messagetype, void *apdu, int apdu_l
 	if (!call || !messagetype || !apdu || (apdu_len < 1) || (apdu_len > 255))
 		return -1;
 
-	new_event = malloc(sizeof(struct apdu_event));
-
-	if (new_event) {
-		memset(new_event, 0, sizeof(struct apdu_event));
-		new_event->message = messagetype;
-		new_event->callback = function;
-		new_event->data = data;
-		memcpy(new_event->apdu, apdu, apdu_len);
-		new_event->apdu_len = apdu_len;
-	} else {
+	if (!(new_event = calloc(1, sizeof(*new_event)))) {
 		pri_error(call->pri, "!! Malloc failed!\n");
 		return -1;
 	}
+
+	new_event->message = messagetype;
+	new_event->callback = function;
+	new_event->data = data;
+	memcpy(new_event->apdu, apdu, apdu_len);
+	new_event->apdu_len = apdu_len;
 	
 	if (call->apdus) {
 		cur = call->apdus;
